/*
 * Copyright (c) 2004-2005 by John Schmerge.
 * All rights reserved.
 *
 * cronfile.c
 */

#if (HAVE_CONFIG_H)
# include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <dirent.h>
#include <unistd.h>
#include <syslog.h>

#include <assert.h>
#include <stdio.h>

#include "libutil/util.h"
#include "cronpaths.h"
#include "cronfile.h"
#include "cronparse.h"
#include "job.h"
#include "cronfile.h"

/* ********************************************************************
 */
#if DEBUG_CRONTAB_READING
void dumpCrontabList(Crontab * list)
{
	Crontab * iter = NULL;

	for (iter = list; iter != NULL; iter = iter->next)
	{
		fprintf(stderr, "\t==>%u\n", iter->uid);
	}
}
#endif

/* ********************************************************************
 */
static void addTabToList(struct Crontab ** list, struct Crontab * new)
{
	struct Crontab * iterator = NULL, * prev = NULL;

	iterator = *list;

	while (iterator != NULL)
	{
		if (iterator->tabOptions.uid > new->tabOptions.uid)
			break;

		assert(iterator->tabOptions.uid != new->tabOptions.uid);

		prev = iterator;
		iterator = iterator->next;
	}

	if (prev == NULL)
	{
		new->next = *list;
		*list = new;
	} else
	{
		new->next = prev->next;
		prev->next = new;
	}
}

/* ********************************************************************
 */
static struct Crontab * removeTabFromList(struct Crontab ** list, uid_t uid)
{
	struct Crontab * iterator = NULL, * prev = NULL, * rc = NULL;

	iterator = *list;

	while (iterator != NULL)
	{
		if (iterator->tabOptions.uid == uid)
			rc = iterator; /* found it */

		if (iterator->tabOptions.uid >= uid)
			break; /* we're done looking */

		prev = iterator;
		iterator = iterator->next;
	}

	/* if found, remove it... prev should be set correctly still */
	if (rc != NULL)
	{
		if (prev == NULL)
		{
			/* handle case where rc is the list head */
			*list = rc->next;
		} else
		{
			/* handle the case where rc is anywhere else in the list */
			prev->next = rc->next;
		}

		/* just to be nice */
		rc->next = NULL;
	}

	return rc;
}

/* ********************************************************************
 */
void destroyTabList(struct Crontab ** list)
{
	struct Crontab * iter = NULL, * tmp = NULL;

	iter = *list;

	while (iter != NULL)
	{
		tmp = iter;
		iter = tmp->next;

		destroyJobList(&tmp->jobList);
		safeFree(tmp);
	}
}

/* ********************************************************************
 */
static void checkSystemTab(struct Crontab * systemTab)
{
	struct stat st;

	if (stat(SYSTEM_CRON_FILE, &st) != 0)
	{
		destroyJobList(&systemTab->jobList);
		systemTab->lastModificationTime = 0;

		return;
	}

	if (st.st_mtime == systemTab->lastModificationTime)
	{
		/* do nothing */
		return;
	}

	/* retire the old tab */
	destroyJobList(&systemTab->jobList);
	systemTab->jobList = NULL;

	/* read the new one */
	if (parseTab(SYSTEM_CRON_FILE, systemTab) == 0)
	{
		destroyJobList(&systemTab->jobList);
		systemTab->jobList = NULL;
	}

	systemTab->lastModificationTime = st.st_mtime;

	systemTab->tabOptions.uid = 0; /* this should always be 0 */
}

/* ********************************************************************
 */
static void checkUserTabs(struct Crontab ** list)
{
	DIR * d = NULL;
	struct dirent * entry = NULL;
	struct passwd * pw = NULL;
	struct stat st;
	char * fname = NULL;
	struct Crontab * newList = NULL, * currentTab = NULL;

	if (stat(CRONTAB_DIR, &st) != 0)
	{
		destroyTabList(list);
		return;
	}

	/* now read the crontabs in the crontab dir */
	if ((d = opendir(CRONTAB_DIR)) != NULL)
	{
		while ((entry = readdir(d)) != NULL)
		{
			/* for simplicity's sake ... */
			fname = entry->d_name;

			/* skip '.' & '..' */
			if ( (fname[0] == '.') &&
			     ( (fname[1] == '\0') ||
			       ( (fname[1] == '.') && (fname[2] == '\0') ) ) )
				continue;

			/* skip non-regular files */
			if ((lstat(fname, &st) != 0) || ( ! S_ISREG(st.st_mode)))
				continue;

			/* skip files that do not correspond to a user in the passwd
			 * file
			 */
			if ((pw = getpwnam(fname)) == NULL)
				continue;

			/* First thing we do is remove the tab from list of crontabs...
			 * this allows us to free it, re-add it or modify it without
			 * having to handle special cases
			 */
			if ((currentTab = removeTabFromList(list, pw->pw_uid)) != NULL)
			{
				if (currentTab->lastModificationTime != st.st_mtime)
				{
					/* free tab, allocate new one & fill it w/ stuff
					 * from file
					 */
					destroyJobList(&currentTab->jobList);
					currentTab->jobList = NULL;
				}
			} else
			{
				/* allocate a new tab */
				currentTab =
				  (struct Crontab *) safeMalloc(sizeof(struct Crontab));

				currentTab->tabOptions.uid = pw->pw_uid;
				currentTab->lastModificationTime = st.st_mtime;
				currentTab->jobList = NULL;
				currentTab->next = NULL;
			}

			if ((currentTab != NULL) && (currentTab->jobList == NULL))
			{
#if DEBUG_CRONTAB_READING
				syslog(LOG_DEBUG, "Parsing Tab file %s", fname);
#endif
				if (parseTab(fname, currentTab) == 0)
				{
					syslog(LOG_INFO, "Parse error occured in the read of "
					       "crontab file %s/%s", CRONTAB_DIR, fname);
					destroyJobList(&currentTab->jobList);
					safeFree(currentTab);
					currentTab = NULL;
				} else
				{
					/* Remember the new date on the file */
					currentTab->lastModificationTime = st.st_mtime;
				}
			}

			if (currentTab != NULL)
				addTabToList(&newList, currentTab);
		}

		closedir(d);
	}

	endpwent();

	destroyTabList(list);

	*list = newList;

#if DEBUG_CRONTAB_READING
	dumpCrontabList(*list);
#endif
}

/* ********************************************************************
 */
void checkCrontabs(struct Crontab * systemTab, struct Crontab ** list)
{
	/* first we attempt the system crontab */
	checkSystemTab(systemTab);

	/* now read the user tabs */
	checkUserTabs(list);

	/* free some mem that was alloc'ed by the readLine func... */
	freeReadlineBuffer();
}

/* ********************************************************************
 * $Log: cronfile.c,v $
 * Revision 1.11  2005/04/25 03:05:41  john
 * Whitespace changes
 *
 * Revision 1.10  2005/04/12 19:56:56  john
 * Fixed bug that causes modified crontabs to not be read
 *
 * Revision 1.9  2005/03/27 01:18:58  john
 * updated copyright
 *
 * Revision 1.8  2005/03/26 01:42:31  john
 * Merge from branch micron_autoconf
 *
 * Revision 1.7  2005/03/26 00:22:32  john
 * Size reduction changes
 *
 * Revision 1.6.2.3  2005/03/25 13:50:44  john
 * Renamed the DEBUG macro to something a little bit more specific
 *
 * Revision 1.6.2.2  2005/03/25 12:56:12  john
 * Size reduction change by using a %s instead of compiler string
 * concatenation in printf statement
 *
 * Revision 1.6.2.1  2005/03/25 02:27:03  john
 * #included config.h and fixed up header code
 *
 * Revision 1.6  2003/05/12 04:03:32  john
 * Added some comments and changed a stat() call to an lstat() call
 *
 * Revision 1.5  2003/02/11 09:16:27  john
 * size-reducing changes; also removed the job input spec
 *
 * Revision 1.4  2003/02/04 07:25:54  john
 * added #include for job.h, as it's no longer implicitly included from
 * cronfile.h
 *
 * Revision 1.3  2003/01/22 07:54:31  john
 * removed users.c & users.h since they're unneeded
 *
 * Revision 1.2  2003/01/22 07:41:49  john
 * removed typedefs in favor of explicit struct defs, made changes
 * to accomodate moving of uid field in Crontab structure.
 *
 * Revision 1.1.1.1  2002/10/06 06:47:18  john
 * initial import of sources
 *
 */
