/*
 * Copyright (c) 2004-2005 by John Schmerge.
 * All rights reserved.
 *
 * job.c
 */

#if (HAVE_CONFIG_H)
# include "config.h"
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>

#include <errno.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>

#include "libutil/util.h"
#include "cronfile.h"
#include "cron.h"
#include "job.h"

#define MINUTESPLIT 30
#define DOW_STAR 0x7F
#define DOM_STAR 0xFFFFFFFE
#define MAX_LOGNAME 256

/* ********************************************************************
 */
static void doLogOpen(int syslogFacility, int options, const char * fmt, ...)
{
	va_list argList;
	static char logName[MAX_LOGNAME];

	va_start(argList, fmt);
	vsnprintf(logName, MAX_LOGNAME, fmt, argList);
	va_end(argList);
	openlog(logName, options, syslogFacility);
}

/* ********************************************************************
 */
static void settupOutput(struct TabOptions * options, int outfd[2])
{
	/* set up output */
	switch (options->loggingType)
	{
	 case LogTypeMail:
	 case LogTypeDevNull:
		outfd[1] = open("/dev/null", O_WRONLY);
		break;

	 /* XXX - we really need to have some mechanism to allow the user to
	  * assign a default mode for newly created files
	  */
	 case LogTypeFileTrunc:
		outfd[1] = open(options->logFileName, O_CREAT|O_WRONLY|O_TRUNC, 0644);
		break;

	 case LogTypeFileAppend:
		outfd[1] = open(options->logFileName, O_CREAT|O_WRONLY|O_APPEND, 0644);
		break;

	 case LogTypeTty:
		outfd[1] = open("/dev/tty", O_WRONLY | O_NOCTTY);
		break;

	 case LogTypeSyslog:
		if (pipe(outfd) != 0)
			outfd[1] = -1; /* do this so that we can deal with the error
			                * in a uniform way below */
		break;
	}

	if (outfd[1] < 0)
	{
		syslog(LOG_ERR, "Couldn't set up job output file descriptors: %m");
		_exit(1);
	}
}

/* ********************************************************************
 */
static pid_t doSyslogging( struct TabOptions * options, int outfd[2] )
{
	pid_t pid = 0;
	FILE * f = NULL;
	char * line = NULL;

	if ((f = fdopen(outfd[0], "r")) == NULL)
	{
		syslog(LOG_ERR, "Couldn't fdopen descriptor %d: %m",
		       outfd[0]);
		_exit(1);
	}

	while ((line = readLine(f)) != NULL)
	{
		syslog(options->syslogPriority, "%s", line);
	}

#if defined(MEM_DEBUG)
	freeReadlineBuffer();
#endif

	return pid;
}

/* ********************************************************************
 */
static void spawnCommand (
	const char * shell,
	struct TabOptions * options,
	struct Job * job )
{
	pid_t pid, loggerPid = -1;
	int childStatus = 0, infd = -1, outfd[2] = {-1, -1};
	char ** env = NULL;

	if ((infd = open("/dev/null", O_RDONLY)) < 0)
	{
		syslog(LOG_ERR, "Couldn't open file %s: %m", "/dev/null");
		_exit(1);
	}

	settupOutput(options, outfd);

	if ((pid = fork()) == 0)
	{
		syslog(LOG_DEBUG, "exec'ing %s", job->command);

		dup2(infd, STDIN_FILENO);
		dup2(outfd[1], STDOUT_FILENO);
		dup2(outfd[1], STDERR_FILENO);

		/* close any open files */
		close(infd);
		close(outfd[0]);
		close(outfd[1]);

		closelog();

		setsid();

		execle(shell, shell, "-c", job->command, NULL, env);

		doLogOpen(LOG_CRON, LOG_PID, "%s(child)", progName);
		syslog(LOG_ERR, "Couldn't exec '%s -c %s': %m", shell, job->command);

		_exit(1);

	} else if (pid == -1)
	{
		/* error */
		syslog(LOG_ERR, "Couldn't fork off child: %m");
		_exit(1);
	}

	close(infd);
	close(outfd[1]);

	/* if we are logging to a pipe rather than a file or device, handle
	 * the situation here
	 */
	if (options->loggingType == LogTypeSyslog)
	{
		doLogOpen(options->syslogFacility, LOG_PID, "%s(uid=%d %s)",
		          progName, options->uid, job->command);
		doSyslogging(options, outfd);
		doLogOpen(LOG_CRON, LOG_PID, "%s(child)", progName);
	} else if (options->loggingType == LogTypeMail)
	{
		/* XXX -- not done */
	}

	close(outfd[0]);

	/* we make sure the command process finished successfully here */
	waitpid(pid, &childStatus, 0);

	/* here we deal with the logger child process */
	if (loggerPid != -1)
		waitpid(loggerPid, &childStatus, 0);
}

/* ********************************************************************
 */
void doCronChild(struct JobQueue * job)
{
	struct passwd * pw = NULL;

	/* get any config things needed */
	if ((pw = getpwuid(job->options->uid)) == NULL)
	{
		syslog(LOG_ERR, "Couldn't find passwd entry for uid %u\n",
		       job->options->uid);
		_exit(1);
	}

	/* we need to de-elevate our priveledges here */
	if (initgroups(pw->pw_name, pw->pw_gid) != 0)
	{
		syslog(LOG_ERR, "Couldn't set up groups for user %s: %m",
		       pw->pw_name);
		_exit(1);
	}

	if (setgid(pw->pw_gid) != 0)
	{
		syslog(LOG_ERR, "Couldn't set gid for user %s: %m",
		       pw->pw_name);
		_exit(1);
	}

	if (setuid(job->options->uid) != 0)
	{
		syslog(LOG_ERR, "Couldn't set uid for child: %m");
		_exit(1);
	}

	/* XXX -- We need a way of setting up env & reading from it */

	/* chdir to user's homedir */
	if (chdir(pw->pw_dir) != 0)
	{
		syslog(LOG_ERR, "Couldn't chdir to %s: %m", pw->pw_dir);
		_exit(1);
	}

	spawnCommand(pw->pw_shell, job->options, job->job);
}

/* ********************************************************************
 */
void processJob(struct JobQueue * job)
{
	pid_t pid;

	if ((pid = fork()) < 0)
	{
		/* XXX we need to revisit the handling of this error */
		syslog(LOG_ERR, "Couldn't fork off child: %m");
	} else if (pid == 0)
	{
		/* do child */
		doLogOpen(LOG_CRON, LOG_PID, "%s(child)", progName);
		doCronChild(job);
		_exit(0);
	}

	/* XXX -- we might want to maintain a list of child processes here for
	 *        intelligent handling of termination signals
	 */
}

/* ********************************************************************
 */
void queueJob(struct JobQueue ** jq, struct TabOptions * opts, struct Job * job)
{
	struct JobQueue * iter = NULL, * newElem = NULL;

	newElem = safeMalloc(sizeof(struct JobQueue));
	newElem->job = job;
	newElem->options = opts;
	newElem->next = NULL;

	if (*jq == NULL)
	{
		*jq = newElem;
		return;
	}

	for (iter = *jq; iter->next != NULL; iter = iter->next)
	{ /* empty */ }

	iter->next = newElem;
}

/* ********************************************************************
 */
void runJobQueue(struct JobQueue ** jq)
{
	struct JobQueue * head = NULL, * tmp = NULL;

	head = *jq;

	while (head != NULL)
	{
#ifdef SIMULATION
		printf("\t(uid = %d) %s\n", head->uid,
		       head->job->command);
#endif
		processJob(head);

		tmp = head;
		head = tmp->next;

		safeFree(tmp);
	}

	*jq = head;
}

/* ********************************************************************
 */
void destroyJobList(struct Job ** list)
{
	struct Job * iterator = NULL, * next = NULL;

	for (next = NULL, iterator = *list; iterator != NULL; iterator = next)
	{
		next = iterator->next;
		safeFree(iterator->command);
		safeFree(iterator);
	}

	*list = NULL;
}

/* ********************************************************************
 */
int jobTimeMatch(struct JobTime * jt, time_t * time)
{
	struct tm * thetime;
	struct JobTime currentJt;

	memset(&currentJt, 0, sizeof(struct JobTime));
	thetime = localtime(time);

	setJobTimeMinute(&currentJt, thetime->tm_min);
	setJobTimeHour(&currentJt, thetime->tm_hour);
	setJobTimeDow(&currentJt, thetime->tm_wday);
	setJobTimeMonth(&currentJt, thetime->tm_mon + 1);
	setJobTimeDom(&currentJt, thetime->tm_mday);

	if ( ! ( (currentJt.minute[0] & jt->minute[0]) ||
	         (currentJt.minute[1] & jt->minute[1]) ) )
	{
		return 0;
	}

	if ( ! (currentJt.hour & jt->hour) )
		return 0;

	if ( ! (currentJt.month & jt->month) )
		return 0;

	/* Ok, this is where stuff gets funky */
	if (((jt->dayOfWeek ^ DOW_STAR) == 0) ||
	    ((jt->dayOfMonth ^ DOM_STAR) == 0))
	{
		if (! ( (currentJt.dayOfWeek & jt->dayOfWeek) &&
		        (currentJt.dayOfMonth & jt->dayOfMonth) ) )
		{
			return 0;
		}
	} else
	{
		if (! ( (currentJt.dayOfWeek & jt->dayOfWeek) ||
		        (currentJt.dayOfMonth & jt->dayOfMonth) ) )
		{
			return 0;
		}
	}

	return 1;
}

/* ********************************************************************
 */
void setJobTimeMinute(struct JobTime * jt, int minute)
{
	if (minute < MINUTESPLIT)
		jt->minute[0] |= (1 << minute);
	else
		jt->minute[1] |= (1 << (minute - MINUTESPLIT));
}

/* ********************************************************************
 */
void setJobTimeHour(struct JobTime * jt, int hour)
{
	jt->hour |= (1 << hour);
}

/* ********************************************************************
   here we avoid bit 0... it complicates our matching logic elsewhere
 */
void setJobTimeDom(struct JobTime * jt, int dom)
{
	jt->dayOfMonth |= (1 << dom);
}

/* ********************************************************************
 */
void setJobTimeMonth(struct JobTime * jt, int month)
{
	jt->month |= (1 << (month - 1));
}

/* ********************************************************************
 */
void setJobTimeDow(struct JobTime * jt, int dow)
{
	jt->dayOfWeek |= (1 << (dow % 7));
}

/* ********************************************************************
 * $Log: job.c,v $
 * Revision 1.14  2005/04/25 03:29:07  john
 * Modified the CronFile struct to contained unsigned char's for enum values
 * instead of int's and added support for specifying logfile name.
 *
 * Revision 1.13  2005/04/25 03:05:41  john
 * Whitespace changes
 *
 * Revision 1.12  2005/04/15 17:27:34  john
 * Size reduction changes, changed the syslogging of job output to not
 * spawn an extra unneeded child. Also removed some of the code that does the
 * logging of child status, as it is handled by the SIGCHLD handler in the
 * same manner.
 *
 * Revision 1.11  2005/03/27 01:20:31  john
 * updated copyright
 *
 * Revision 1.10  2005/03/26 01:42:31  john
 * Merge from branch micron_autoconf
 *
 * Revision 1.9  2005/03/26 00:24:31  john
 * Size reduction changes
 *
 * Revision 1.8.2.1  2005/03/25 02:27:03  john
 * #included config.h and fixed up header code
 *
 * Revision 1.8  2003/07/09 04:32:06  john
 * Bugfix by Frank Geerlings for lack of a setgid call in child
 * process.
 *
 * Revision 1.7  2003/07/09 04:20:07  john
 * Fixed spelling error
 *
 * Revision 1.6  2003/04/27 23:08:10  john
 * string table size reduction change
 *
 * Revision 1.5  2003/02/11 09:16:27  john
 * size-reducing changes; also removed the job input spec
 *
 * Revision 1.4  2003/02/10 05:48:41  john
 * Small ammount of code refactoring
 *
 * Revision 1.3  2003/02/09 20:46:38  john
 * added job output support; everything but mailing of job output works
 * correctly
 *
 * Revision 1.2  2003/01/22 07:49:46  john
 * removed typedefs
 *
 * Revision 1.1.1.1  2002/10/06 06:47:18  john
 * initial import of sources
 *
 */
