/*
 * Copyright (c) 2003-2005 by John Schmerge
 * All rights reserved.
 *
 * util.c
 */

#if defined(HAVE_CONFIG_H)
# include "config.h"
#endif

#include <sys/types.h>
#include <unistd.h>

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <syslog.h>

#include "util.h"

#define DEFAULT_RESULT_SIZE 256

static int result_size = 0;
static unsigned char * result = NULL;

#if (MEM_DEBUG)
int transaction_num = 0;
#endif

#if (TRACE_OVERWRITES)
static const size_t tracer1 = 0xAA55;
static const size_t tracer2 = 0x55AA;
#endif

static MallocHandler theMallocHandler = defaultMallocHandler;

/* **************************************************************** */
void defaultMallocHandler(void)
{
	abort();
}

/* **************************************************************** */
MallocHandler setMallocHandler(MallocHandler h)
{
	MallocHandler rc = theMallocHandler;

	theMallocHandler = h;

	return rc;
}


/* **************************************************************** */
char * safeStrdup(const char * s)
{
	char * tmp = NULL;

	assert(s != NULL);

	tmp = (char *) safeMalloc(strlen(s) + 1);
	strcpy(tmp, s);

	return tmp;
}

/* **************************************************************** */
#if (MEM_DEBUG)
void * _safeMalloc(size_t sz, int line, char * file)
#else
void * _safeMalloc(size_t sz)
#endif
{
	void * ptr = NULL;
	size_t alloc_sz = sz;

#if (ZERO_FREED_MEM)
# if (TRACE_OVERWRITES)
	alloc_sz += (sizeof(size_t) * 6);
# else
	alloc_sz += sizeof(size_t);
# endif
#endif

	if ((ptr = malloc(alloc_sz)) == NULL)
		theMallocHandler();

#if (MEM_DEBUG)
	fprintf(stderr, "A %d %p %d %d bytes at line %d in file %s\n",
	        getpid(), ptr, ++transaction_num, alloc_sz, line, file);
#endif

	memset(ptr, 0, alloc_sz);

#if (ZERO_FREED_MEM)
# if (TRACE_OVERWRITES)
	memcpy(ptr, &tracer1, sizeof(size_t));
	memcpy(ptr + sizeof(size_t), &sz, sizeof(size_t));
	memcpy(ptr + (2 *sizeof(size_t)), &tracer2, sizeof(size_t));
	ptr += (3 * sizeof(size_t));

	memcpy(ptr + sz, &tracer2, sizeof(size_t));
	memcpy(ptr + sz + sizeof(size_t), &sz, sizeof(size_t));
	memcpy(ptr + sz + (2 *sizeof(size_t)), &tracer1, sizeof(size_t));
# else
	memcpy(ptr, &sz, sizeof(size_t));
	ptr += sizeof(size_t);
# endif
#endif

	return ptr;
}

/* **************************************************************** */
void safeFree(void * ptr)
{
#if (ZERO_FREED_MEM)
	size_t sz = 0;
# if (TRACE_OVERWRITES)
	size_t myTracer1 = 0, myTracer2 = 0, sz2 = 0;
# endif
#endif

	if (ptr != NULL)
	{
#if (ZERO_FREED_MEM)
# if (TRACE_OVERWRITES)
		ptr -= sizeof(size_t);
		memcpy(&myTracer2, ptr, sizeof(size_t));
		assert(myTracer2 == tracer2);

		ptr -= sizeof(size_t);
		memcpy(&sz, ptr, sizeof(size_t));

		ptr -= sizeof(size_t);
		memcpy(&myTracer1, ptr, sizeof(size_t));
		assert(myTracer1 == tracer1);

		memcpy(&myTracer2, ptr + sz + (3 * sizeof(size_t)), sizeof(size_t));
		assert(myTracer2 == tracer2);

		memcpy(&sz2, ptr + sz + (4 * sizeof(size_t)), sizeof(size_t));
		assert(sz2 == sz);

		memcpy(&myTracer1, ptr + sz + (5 * sizeof(size_t)), sizeof(size_t));
		assert(myTracer1 == tracer1);
		memset(ptr, 0, sz + (6 *sizeof(size_t)));
# else
		ptr -= sizeof(size_t);
		memcpy(&sz, ptr, sizeof(size_t));
		memset(ptr, 0, sz + sizeof(size_t));
# endif
#endif

#if (MEM_DEBUG)
		fprintf(stderr, "F %d %p %d\n", getpid(), ptr, ++transaction_num);
#endif
		free(ptr);
	}
}

/* **************************************************************** */
unsigned char * readLine(FILE * f)
{
	int i = 0, c = 0;
	unsigned char * tmp = NULL;

	if (result == NULL)
	{
		result_size = DEFAULT_RESULT_SIZE;
		result = (unsigned char *) safeMalloc(result_size);
	}

	do {
		clearerr(f);
		errno = 0;

		while (((c = getc(f)) != -1) && (c != '\n'))
		{
			if (i >= result_size)
			{
				/* double the buffer size */
				tmp = (unsigned char *) safeMalloc(result_size * 2);
				memcpy(tmp, result, result_size);
				safeFree(result);
				result = tmp;
				result_size *= 2;
				tmp = NULL;
			}

			result[i] = c;

			++i;
		}
	} while (ferror(f) && (errno == EINTR));

	result[i] = '\0';

	if (i == 0 && feof(f))
		return NULL;
	else
		return result;
}

/* **************************************************************** */
void freeReadlineBuffer(void)
{
	safeFree(result);

	result = NULL;
	result_size = DEFAULT_RESULT_SIZE;
}

/* ********************************************************************
 * $Log: util.c,v $
 * Revision 1.7  2005/04/25 03:05:43  john
 * Whitespace changes
 *
 * Revision 1.6  2005/04/12 12:34:29  john
 * Merge from DEVEL branch
 *
 * Revision 1.1.1.1.2.6  2005/04/12 12:28:53  john
 * Added pid to information spit out in memdebug mode
 *
 * Revision 1.5  2005/04/12 03:13:49  john
 * Pulled new libutil work into project
 *
 * Revision 1.1.1.1.2.5  2005/04/12 03:02:58  john
 * Cleaned up safeFree routine and fixed core dump
 *
 * Revision 1.1.1.1.2.4  2005/04/12 02:57:11  john
 * Moved location of allocation trace statement to a place where it prints
 * out the actual base address of the mem allocated
 *
 * Revision 1.1.1.1.2.3  2005/04/12 02:42:46  john
 * Added heap overwrite protection
 *
 * Revision 1.1.1.1.2.2  2005/04/12 02:22:54  john
 * Merge from head
 *
 * Revision 1.4  2005/03/27 01:21:09  john
 * updated copyright
 *
 * Revision 1.3  2005/03/26 01:42:34  john
 * Merge from branch micron_autoconf
 *
 * Revision 1.2  2005/03/26 00:26:14  john
 * Removing the indent directive for size reduction
 *
 * Revision 1.1.1.1.4.1  2005/03/25 02:22:04  john
 * Modified cpp conditionals to play nice with autoconf
 *
 * Revision 1.1.1.1.2.1  2005/03/17 12:22:46  john
 * added memory overwrite protection
 *
 * Revision 1.1.1.1  2005/02/26 12:57:27  john
 * initial import of sources
 */
